import numpy as  np
from collections import deque

##################################################################################
def otimes(container):
    '''Usage: c = otimes((a,b))'''
    return(np.sum(container))

def oplus(container):
    '''Usage: c = oplus((a,b))'''
    m = np.amin(container)
    if m==np.inf:
        return(np.inf)
    else:
        return(m-np.log(np.sum(np.exp(m-np.array(container)))))

def todo_lexicon2wfst(filename):
    '''
    Inputs:
    filename = a string, giving the name of a file that should be read as a lexicon.
      Each line is: word phone1 phone2 ...
      You will need to split the phones, and create an transition for each.
      CAUTION: treat multiple spaces as a single space, i.e., 'a  b' -> 'a','b'
    Outputs:
    L = a list of transitions, representing a deterministic FST, with transitions in order of the lexicon.
      Each transition is a 5-tuple: (previous state, input string, output string, weight=0, next state).
      All states should be non-negative integers, beginning with state 0, created in the order
      that they are needed in order to create the lexical entries in order.
      Every entry in the lexicon should be a path from state 0, back to state 0,
      by way of N other states, where N is the number of phones in the word.
      The first N transitions in each word should have phones as input strings, and '' as output string.
      The last transition in each word should have '' as input string, and the word as output string.
      Weights should all be 0.
    Lfinal = a list of the final states of L.
    '''
    L = []
    Lfinal = []
    raise NotImplementedError('You need to write this part!')
    return(L,Lfinal)
                              
def todo_unigram(textfile, L):
    '''
    Inputs:
    textfile = a string, giving the name of a file that should be read as language model training text.
      CAUTION: treat multiple spaces as a single space, i.e., 'a  b' -> 'a','b'
    L = a list of WFST transitions, specifying all words in the lexicon
    Outputs:
    G = a list of transitions, specifying a unigram grammar.
      Each transition is a 5-tuple: (previous state=0, ilabel, olabel=ilabel, weight, next state=0).
      There is only one state: state 0.
      ilabel and olabel are the same; they both equal the word.  
      There should be one transition for each word in the lexicon.
      The weight should be the negative log probability of that word, estimated based on its
      frequency in textfile, Laplace-smoothed with a smoothing factor of 1.
    Gfinal = a list of the final states of G.
    '''
    G = []
    Gfinal = []
    raise NotImplementedError('You need to write this part!')
    return(G, Gfinal)

def todo_transcript2wfst(filename):
    '''
    Inputs:
    filename = a string, giving the name of a file that should be read as an IPA phone transcript.
      CAUTION: treat multiple spaces as a single space, i.e., 'a  b' -> 'a','b'
    Outputs:
    T = a list of transitions.
      Each transition is a 5-tuple: (previous state, ilabel, olabel=ilabel, weight=0, next state).
      Each ilabel and olabel are the same, they equal the phone symbol.
      There should be as many transitions as there are phone symbols in the input file.
      They should proceed from state 0 to state 1, to state 2, and so on in order.
    Tfinal = a list of the final states of T.
    '''
    T = []
    Tfinal = []
    raise NotImplementedError('You need to write this part!')
    return(T, Tfinal)

def transitions2states(E):
    '''Find all unique state IDs referenced in a list of  transitions'''
    Q = set()
    for t in E:
        Q.add(t[0])
        Q.add(t[-1])
    return(sorted(list(Q)))

def cross_product(QA,QB):
    '''
    Usage: QCdef = cross_product(QA,QB).
    QA, QB = lists of states in A and B.
    QCdef = dict of dicts such that QCdef[qa][qb] = qc,
    where qc = qb+len(QB)*qa.
    '''
    QCdef = {}
    for qa in QA:
        QCdef[qa]={}
        for qb in QB:
            QCdef[qa][qb] = qb+len(QB)*qa
    return(QCdef)

def todo_fstcompose(A, Afinal, B, Bfinal):
    '''
    Inputs:
    A: a list of WFST transitions
    Afinal: list of the final states in A
    B: a list of WFST transitions
    Bfinal: list of the final states in B
    Output:
    C: a list of WFST transitions, composition of A and B, with state indices
       that are the cross_product of A's states and B's states.  Transitions  should be
       sorted in order of increasing previous state, subsorted in order of increasing next state.
    Cfinal: list of the final states in C, sorted
    '''
    C = []
    Cfinal = []
    raise NotImplementedError('You need to write this part!')
    return(C, Cfinal)

def todo_sort_topologically(A,Afinal):
    '''
    Inputs:
    A: an FST that contains no cycles
    Afinal: a list of the final states in A
    Outputs:
    B: the same FST, with states renumbered so that n[t]>=p[t] for every transition.
    Bfinal: a  list of the final states in B
    '''
    B = []
    Bfinal = []
    raise NotImplementedError('You need to write this part!')
    return(B,Bfinal)
    
def todo_fstbestpath(TLG,TLGfinal):
    '''
    Inputs:
    TLG: the composed FST, topologically sorted
    TLGfinal: the list of allowed final states
    Outputs: 
    delta: a map from states to delta surprisals
    psi: a map from states to the best preceding transition
      Tie-breaker: if two transitions into q have the same cost, psi[q] should point to
      the one tested first (the one sorted earlier by todo_sort_topologically).
    bestpath: the sequence of transitions forming the best path, in order from start to finish
    '''
    TLG = sorted(TLG, key = lambda transition: transition[0]) # sort by previous state
    delta = { q:np.inf for q in transitions2states(TLG) }
    delta[0] = 0
    psi = {}
    bestpath = []
    raise NotImplementedError('You need to write this part!')
    return(delta, psi, bestpath[::-1])

def todo_fstforward(TLG):
    '''
    Inputs:
    TLG: a list of  WFST transitions, with states sorted topologically
    Output:
    alpha: forward  surprisals of each state in TLG
    '''
    TLG = sorted(TLG, key = lambda transition: transition[0]) # sort by previous state
    alpha = { q:np.inf for q in transitions2states(TLG) }
    alpha[0] = 0
    raise NotImplementedError('You need to write this part!')
    return(alpha)

def todo_fstbackward(TLG, TLGfinal):
    '''
    Inputs:
    TLG: a list of  WFST transitions, with states sorted topologically
    TLGfinal: a list of the final states of TLG
    Output:
    beta: backward surprisals of each state in TLG
    '''
    TLG = sorted(TLG, key = lambda transition: transition[-1]) # sort by next state
    beta = { q:np.inf for q in transitions2states(TLG) }
    for q in TLGfinal:
        beta[q] = 0
    raise NotImplementedError('You need to write this part!')
    return(beta)

##############################################################################
def fstreestimate(TLG, L, Lfinal, alpha, beta):
    '''
    Inputs:
    TLG: a list of transitions, with state indices matching alpha and beta
    L: a list of transitions, with input and output labels matching TLG
    Lfinal: a list of the final states of L
    alpha: forward surprisal of each state in TLG
    beta: backward surprisal of each state in TLG
    Output:
    LG: a copy of L, with all of its transition weights re-estimated.
      Transition tL with input label i[tL] and output label o[tL]
      is re-estimated as the oplus, over all transitions tTLG that have the 
      same i and o labels, of alpha[p[tTLG]] otimes w[tTLG] otimes beta[n[tTLG]].
    '''
    #
    # First, find xi[istr,ostr], the expected number of translations of istr and ostr
    Sigma = set([ tlg[1] for tlg in L+TLG ])
    Omega = set([ tlg[2] for tlg in L+TLG ])
    xi = { istr:{ ostr:np.inf for ostr in Omega } for istr in Sigma }
    for tlg in TLG:
        xi[tlg[1]][tlg[2]] = oplus((xi[tlg[1]][tlg[2]], otimes((alpha[tlg[0]],tlg[3],beta[tlg[-1]]))))
    #
    # Second, accumulate the transition-probability denominators for each state in L
    gamma = { q:np.inf for q in transitions2states(L) }
    for t in L:
        gamma[t[0]] = oplus((gamma[t[0]], xi[t[1]][t[2]]))
    #
    # Third, calculate the state-conditional transition surprisals
    LG = []
    for t in L:
        if xi[t[1]][t[2]] < np.inf:
            LG.append((t[0],t[1],t[2], otimes((xi[t[1]][t[2]], -gamma[t[0]])),t[-1]))
        else:
            LG.append((t[0],t[1],t[2], np.inf, t[-1]))
    LGfinal = Lfinal.copy()
    return(LG,LGfinal)

##############################################################################
if __name__=="__main__":
    import argparse
    parser = argparse.ArgumentParser('Run MP5 to generate results.json.')
    parser.add_argument('-L','--lexicon',default='data/lexicon.txt',
                        help='''lexicon.  Default: "data/lexicon.txt"''')
    parser.add_argument('-G','--languagemodeltexts',default='data/languagemodeltexts.txt',
                        help='''language model texts.  Default: "data/languagemodeltexts.txt"''')
    parser.add_argument('-T','--transcript',default='data/transcript.txt',
                        help='''phonetic transcript.  Default: "data/transcript.txt"''')
    args = parser.parse_args()

    T, Tfinal =todo_transcript2wfst(args.transcript)
    L, Lfinal = todo_lexicon2wfst(args.lexicon)
    G, Gfinal = todo_unigram(args.languagemodeltexts,L)
    LG, LGfinal = todo_fstcompose(L,Lfinal,G,Gfinal)
    TLG, TLGfinal = todo_fstcompose(T,Tfinal,LG,LGfinal)
    TLG_sorted, TLGfinal_sorted = todo_sort_topologically(TLG,TLGfinal)
    delta, psi, bestpath = todo_fstbestpath(TLG_sorted,TLGfinal_sorted)
    alpha = todo_fstforward(TLG_sorted)
    beta = todo_fstbackward(TLG_sorted, TLGfinal_sorted)

    import json
    with open('results.json','w')  as f:
        data={'T':T,'Tfinal':Tfinal,'L':L,'Lfinal':Lfinal,'G':G,'Gfinal':Gfinal,
              'LG':LG,'LGfinal':LGfinal,'TLG':TLG,'TLGfinal':TLGfinal,
              'TLG_sorted':TLG_sorted,'TLGfinal_sorted':TLGfinal_sorted,
              'bestpath':bestpath,'alpha':alpha,'beta':beta}
        json.dump(data,f)
